"""Query helper utilities for executing SQL against Apache Iceberg.

This module provides a small helper class to execute SQL queries and return
results as a list of dictionaries.
"""

from pathlib import Path
from typing import Annotated, Optional

from duckdb import DuckDBPyConnection
from duckdb import connect as ddb_connect
from pyarrow.csv import CSVWriter
from pyarrow.ipc import RecordBatchFileWriter
from pyarrow.parquet import ParquetWriter
from pydantic import Field
from pyiceberg.catalog import Catalog, CatalogType, infer_catalog_type


class QueryTools:
    """Utilities for executing SQL queries against Iceberg via a DuckDB connection.

    Attributes:
        duckdb: Active DuckDB connection.
    """

    duckdb: DuckDBPyConnection

    def __init__(self, duckdb: DuckDBPyConnection) -> None:
        """Initialize QueryTools with a DuckDB connection.

        Args:
            connection: The DuckDB connection to use
                for executing queries.
        """
        self.duckdb = duckdb

    async def sql_query(
        self,
        query: Annotated[str, Field(description="DuckDB SQL query.")],
        file: Annotated[Optional[Path], Field(description="File path to write SQL query results.")] = None,
    ) -> Annotated[str, Field(description="Query results by row as JSON.")]:
        """Execute a SQL query and return results as a list of dicts or write to a file.

        The query is executed using the embedded DuckDB connection and the
        result is materialized into a Polars DataFrame. If no file is provided,
        the result is converted to a list of dictionaries and returned as JSON.
        If a file is provided, the results are written to the specified file in the
        appropriate format (CSV, Parquet, or Feather).

        Note:
            When querying Iceberg tables, the SQL table name should be of the format catalog.table_identifier.

        Args:
            query: The SQL query string to execute.
            file: Optional file path to write SQL query results. If provided,
                the results will be written to the specified file instead of being returned as JSON.

        Returns:
            Query results by row as JSON if file is not provided, otherwise a message
            indicating the file path and size of the written query results.

        Raises:
            FileNotFoundError: If the parent directory of the specified file does not exist.
            ValueError: If the file extension is unsupported.
        """
        result = self.duckdb.sql(query).execute()

        if file is None:
            return result.pl().write_json()
        elif file.parent.is_dir() is False:
            raise FileNotFoundError(f"Parent directory {file.parent.resolve()} must exist!")
        else:
            batch_reader = result.arrow()
            match file.suffix:
                case ".csv":
                    writer = CSVWriter(file, batch_reader.schema)
                case ".parquet" | ".pqt":
                    writer = ParquetWriter(file, batch_reader.schema)
                case ".feather" | ".ftr":
                    writer = RecordBatchFileWriter(file, batch_reader.schema)
                case _:
                    raise ValueError(f"Unsupported file extension: {file.suffix}")

            list(map(writer.write_batch, batch_reader))
            writer.close()

            return f"Query result file: {file.resolve()} has file size of {file.stat().st_size} bytes."


def load_duckdb(catalog: Catalog) -> Optional[DuckDBPyConnection]:
    """Create and configure a DuckDB connection with the Iceberg extension.

    The function connects to an in-memory DuckDB instance, loads the
    Iceberg extension, and attaches an Iceberg catalog using the same
    environment variables as :func:`load_catalog`.

    Returns:
        The configured DuckDB connection.
    """
    con = ddb_connect()
    con.load_extension("iceberg")

    catalog_type = infer_catalog_type(catalog.name, catalog.properties)

    match catalog_type:
        case CatalogType.GLUE:
            con.load_extension("aws")
            if "glue.profile-name" in catalog.properties:
                # Glue Catalog using AWS profile
                con.sql(f"""
                    CREATE OR REPLACE SECRET (
                        TYPE s3,
                        PROVIDER credential_chain,
                        CHAIN config,
                        PROFILE '{catalog.properties["glue.profile-name"]}'
                        REGION '{catalog.properties["glue.region"]}'
                    );
                    """)
            elif "glue.access-key-id" in catalog.properties:
                # Glue Catalog using configured credentials
                con.sql(f"""
                        CREATE OR REPLACE SECRET (
                            TYPE s3,
                            PROVIDER config,
                            KEY_ID '{catalog.properties["glue.access-key-id"]}',
                            SECRET '{catalog.properties["glue.secret-access-key"]}',
                            REGION '{catalog.properties["glue.region"]}'
                        );
                        """)
            else:
                return None
            con.sql(f"""
                    ATTACH '{catalog.properties["glue.id"]}' AS catalog (
                    TYPE iceberg,
                    ENDPOINT_TYPE 'glue'
                    );
                    """)
        case CatalogType.REST:
            if "oauth2-server-uri" in catalog.properties:
                # REST Catalog using OAuth
                con.sql(f"""
                        CREATE OR REPLACE SECRET (
                            TYPE iceberg,
                            CLIENT_ID '{catalog.properties["client-id"]}',
                            CLIENT_SECRET '{catalog.properties["client-secret"]}',
                            OAUTH2_SERVER_URI '{catalog.properties["oauth2-server-uri"]}'
                        );
                        """)
                con.sql(f"""
                        ATTACH '{catalog.properties.get("warehouse", "")}' AS catalog (
                            TYPE iceberg,
                            ENDPOINT '{catalog.properties["uri"]}'
                        );
                        """)
            elif "s3tablescatalog" in catalog.properties.get("warehouse", ""):
                # S3 Tables Catalog
                con.load_extension("aws")
                con.sql("""
                        CREATE OR REPLACE SECRET (
                            TYPE s3,
                            PROVIDER credential_chain
                        );
                        """)
                con.sql(f"""
                        ATTACH '{catalog.properties["warehouse"]}' AS catalog (
                        TYPE iceberg,
                        ENDPOINT_TYPE s3_tables
                        );
                        """)

            else:
                # REST Catalog using token
                con.sql(f"""
                        CREATE SECRET (
                            TYPE iceberg,
                            TOKEN '{catalog.properties["token"]}'
                        );
                        """)
                con.sql(f"""
                        ATTACH '{catalog.properties.get("warehouse", "")}' AS catalog (
                            TYPE iceberg,
                            ENDPOINT '{catalog.properties["uri"]}'
                        );
                        """)

        case _:
            return None

    return con



"""Table-related helpers wrapping a PyIceberg Catalog.

This module exposes TableTools, a convenience wrapper around a Catalog
to list tables, read table metadata, and page table contents into Python
dictionaries.
"""

from pathlib import Path
from typing import Annotated, Dict, List, Literal, Optional, Union

import polars as pl
from pyarrow import Table
from pyarrow.csv import CSVWriter
from pyarrow.ipc import RecordBatchFileWriter
from pyarrow.parquet import ParquetWriter
from pydantic import Field
from pyiceberg.catalog import Catalog
from pyiceberg.table.metadata import TableMetadata
from pyiceberg.table.snapshots import Snapshot
from pyiceberg.typedef import Identifier


class TableTools:
    """Convenience helpers for reading Iceberg table information.

    Attributes:
        catalog: Catalog instance used to load
            and inspect tables.
    """

    catalog: Catalog

    def __init__(self, catalog: Catalog) -> None:
        """Initialize TableTools.

        Args:
            catalog: The catalog used to access
                tables.
        """
        self.catalog = catalog

    async def list_tables(
        self,
        namespace: Annotated[Union[str, Identifier], Field(description="The namespace to list tables from.")],
    ) -> Annotated[List[Identifier], Field(description="List of table identifiers in namespace.")]:
        """List all tables in a namespace.

        Args:
            namespace: The namespace to list tables from.

        Returns:
            A list of table identifiers in the namespace.
        """

        return self.catalog.list_tables(namespace)

    async def read_table_metadata(
        self,
        identifier: Annotated[Union[str, Identifier], Field(description="The identifier of the table.")],
    ) -> Annotated[TableMetadata, Field(description="Table metadata object.")]:
        """Retrieve metadata for a table.

        Args:
            identifier: The identifier of the table.

        Returns:
            The table metadata object.
        """
        table = self.catalog.load_table(identifier)

        return table.metadata

    async def read_table_snapshots(
        self,
        identifier: Annotated[Union[str, Identifier], Field(description="The identifier of the table.")],
        snapshot_id: Annotated[Optional[int], Field(description="Optional snapshot ID for time travel.")] = None,
        limit: Annotated[Optional[int], Field(description="Maximum number of snapshots to return.")] = None,
    ) -> Annotated[List[Snapshot], Field(description="List of snapshot objects.")]:
        """
        Retrieve snapshot information for a table.

        Args:
            identifier: The identifier of the table.
            snapshot_id: Optional snapshot ID for time travel queries.
            limit: Optional limit on number of snapshots to return.

        Returns:
            List of snapshot objects.
        """
        table = self.catalog.load_table(identifier)
        snapshots = table.inspect.snapshots()

        if snapshot_id is not None:
            snapshots = [s for s in snapshots if s.snapshot_id == snapshot_id]

        if limit is not None:
            snapshots = snapshots[:limit]

        return snapshots

    async def read_table_contents(
        self,
        identifier: Annotated[Union[str, Identifier], Field(description="The identifier of the table.")],
        start: Annotated[
            int,
            Field(description="Row index to start pagination, inclusive."),
        ] = 0,
        end: Annotated[
            Optional[int],
            Field(description="Row index to end pagination, exclusive."),
        ] = None,
    ) -> Annotated[str, Field(description="JSON representation of the table rows.")]:
        """Retrieve table contents with optional pagination.

        Supports negative indices for both `start` and `end` parameters to enable
        reading from the end of the table. When ``end`` is ``None`` (default),
        all rows from ``start`` to the end of the table are returned.

        Args:
            identifier: The identifier of the table.
            start: Row index to start pagination (inclusive). Defaults to
                ``0``. Negative indices count from the end of the table.
            end: Row index to end pagination (exclusive).
                Defaults to ``None`` (end of table). Negative indices count
                from the end of the table.

        Returns:
            JSON representation of the table rows.
        """
        table = self.catalog.load_table(identifier)

        snapshot = table.current_snapshot()
        if snapshot is None:
            raise ValueError(f"Table: {identifier} has no current snapshot.")
        summary = snapshot.summary
        if summary is None:
            raise ValueError(f"Snapshot for Table: {identifier} has no summary.")
        row_limit = int(summary.get("total-records", "0"))

        if start < 0:
            start += row_limit
        if end is None:
            end = row_limit
        elif end < 0:
            end = max(0, end + row_limit)
        else:
            end = min(end, row_limit)

        try:
            df = table.to_polars()
            df = await df.slice(start, end - start).collect_async()

        except OSError:
            df = table.scan().to_polars()
            df = df.slice(start, end - start)

        return df.write_json()

    async def download_table_contents(
        self,
        identifier: Annotated[Union[str, Identifier], Field(description="The identifier of the table.")],
        file: Annotated[Path, Field(description="Path of downloaded table file.")],
    ) -> None:
        """Download table contents to a file.

        Args:
            identifier: The identifier of the table.
            file: Path of downloaded table file.

        Raises:
            FileNotFoundError: If the parent directory does not exist.
            ValueError: If the file extension is unsupported.
        """
        if file.parent.is_dir() is False:
            raise FileNotFoundError(f"Parent directory {file.parent.resolve()} must exist!")

        table = self.catalog.load_table(identifier)
        batch_reader = table.scan().to_arrow_batch_reader()

        match file.suffix:
            case ".csv":
                writer = CSVWriter(file, batch_reader.schema)
            case ".parquet" | ".pqt":
                writer = ParquetWriter(file, batch_reader.schema)
            case ".feather" | ".ftr":
                writer = RecordBatchFileWriter(file, batch_reader.schema)
            case _:
                raise ValueError(f"Unsupported file extension: {file.suffix}")

        list(map(writer.write_batch, batch_reader))
        writer.close()

    async def create_table(
        self,
        identifier: Annotated[Union[str, Identifier], Field(description="The identifier of the table.")],
        contents: Annotated[
            Optional[Dict[str, List]], Field(description="Columnar dictionary of table contents.")
        ] = None,
        file: Annotated[Optional[Path], Field(description="Path to table file.")] = None,
    ) -> Annotated[str, Field(description="JSON representation of last 5 table rows.")]:
        """Create a new Iceberg table and populate it with contents.

        Creates a new table in the catalog with the specified identifier using
        the schema inferred from the provided contents/file, then overwrites the table
        with the actual data. Either contents or file must be provided.

        Args:
            identifier: The identifier of the table to create.
            contents: A columnar dictionary where keys are column names and values
                are lists of column data.
            file: Path to table file. All file types supported by Polars can be used.

        Raises:
            ValueError: If none or both contents and file are provided.
        """

        if contents is not None and file is not None:
            raise ValueError("Only one of contents or file can be provided!")
        elif contents is not None:
            table_contents = Table.from_pydict(contents)
        elif file is not None:
            table_contents = self._read_table_from_file(file)
        else:
            raise ValueError("One of contents or file must be provided!")

        table = self.catalog.create_table(identifier, table_contents.schema)

        table.overwrite(table_contents)

        return await self.read_table_contents(identifier, start=-5)

    async def write_table(
        self,
        identifier: Annotated[Union[str, Identifier], Field(description="The identifier of the table.")],
        mode: Annotated[
            Literal["append", "overwrite"], Field(description="Append the contents or overwrite the table.")
        ],
        contents: Annotated[
            Optional[Dict[str, List]], Field(description="Columnar dictionary of table contents.")
        ] = None,
        file: Annotated[Optional[Path], Field(description="Path to table file.")] = None,
    ) -> Annotated[str, Field(description="JSON representation of last 5 table rows.")]:
        """Write data to an existing Iceberg table.

        Args:
            identifier: The identifier of the table.
            mode: Append the contents or overwrite the table.
            contents: A columnar dictionary where keys are column names and values
                are lists of column data.
            file: Path to table file. All file types supported by Polars can be used.

        Returns:
            JSON representation of last 5 table rows.

        Raises:
            ValueError: If both contents and file are provided, or if neither is provided,
                or if an invalid write mode is provided.
        """
        if contents is not None and file is not None:
            raise ValueError("Only one of contents or file can be provided!")
        elif contents is not None:
            table_contents = Table.from_pydict(contents)
        elif file is not None:
            table_contents = self._read_table_from_file(file)
        else:
            raise ValueError("One of contents or file must be provided!")

        table = self.catalog.load_table(identifier)

        match mode:
            case "append":
                table.append(table_contents)
            case "overwrite":
                table.overwrite(table_contents)
            case _:
                raise ValueError(f"Invalid write table mode provided: {mode}")

        return await self.read_table_contents(identifier, start=-5)

    async def delete_table(
        self, identifier: Annotated[Union[str, Identifier], Field(description="The identifier of the table.")]
    ) -> Annotated[List[Identifier], Field(description="List of remaining tables in namespace.")]:
        """Delete a table from the catalog.

        Args:
            identifier: The identifier of the table.

        Returns:
            List of remaining tables in namespace.
        """
        self.catalog.drop_table(identifier)

        namespace = Catalog.namespace_from(identifier)

        return await self.list_tables(namespace)

    def _read_table_from_file(self, file: Annotated[Path, Field(description="Path of table file.")]) -> Table:
        """Read table contents from a file.

        Args:
            file: Path of table file.

        Returns:
            PyArrow Table containing the file contents.

        Raises:
            ValueError: If the file extension is unsupported.
        """
        match file.suffix:
            case ".csv":
                return pl.read_csv(file).to_arrow()
            case ".xslx" | ".xls":
                return pl.read_excel(file).to_arrow()
            case ".json":
                return pl.read_json(file).to_arrow()
            case ".parquet" | ".pqt":
                return pl.read_parquet(file).to_arrow()
            case ".avro":
                return pl.read_avro(file).to_arrow()
            case ".feather" | ".ftr":
                return pl.read_ipc(file).to_arrow()
            case _:
                raise ValueError(f"Unsupported file extension: {file.suffix}")



from unittest import IsolatedAsyncioTestCase
from unittest.mock import Mock

from pyiceberg.catalog import Catalog

from src.tools.namespace import NamespaceTools


class TestNamespace(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.mock_catalog = Mock(spec=Catalog)
        self.tools = NamespaceTools(self.mock_catalog)

    async def test_list_namespaces_returns_all_namespaces(self) -> None:
        namespace_list = [("namespace1",), ("namespace2",)]
        self.mock_catalog.list_namespaces.return_value = namespace_list

        result = await self.tools.list_namespaces()

        self.assertEqual(result, namespace_list)
        self.mock_catalog.list_namespaces.assert_called_once()

    async def test_list_namespaces_with_parent_returns_child_namespaces(self) -> None:
        child_namespace_list = [("child_namespace",)]
        self.mock_catalog.list_namespaces.return_value = child_namespace_list

        result = await self.tools.list_namespaces("parent_namespace")

        self.assertEqual(result, child_namespace_list)
        self.mock_catalog.list_namespaces.assert_called_once_with("parent_namespace")

    async def test_create_namespace_creates_and_returns_updated_list(self) -> None:
        new_namespace_list = [("new_namespace",), ("existing_namespace",)]
        self.mock_catalog.create_namespace_if_not_exists.return_value = None
        self.mock_catalog.list_namespaces.return_value = new_namespace_list

        result = await self.tools.create_namespace("new_namespace")

        self.mock_catalog.create_namespace_if_not_exists.assert_called_once_with("new_namespace")
        self.assertEqual(result, new_namespace_list)

    async def test_delete_namespace_drops_and_returns_updated_list(self) -> None:
        remaining_namespace_list = [("remaining_namespace",)]
        self.mock_catalog.drop_namespace.return_value = None
        self.mock_catalog.list_namespaces.return_value = remaining_namespace_list

        result = await self.tools.delete_namespace("test_namespace")

        self.mock_catalog.drop_namespace.assert_called_once_with("test_namespace")
        self.assertEqual(result, remaining_namespace_list)



"""Namespace management helpers that wrap a PyIceberg Catalog.

This module exposes NamespaceTools, a thin helper around a Catalog
instance to list and create namespaces.
"""

from typing import Annotated, List, Union

from pydantic import Field
from pyiceberg.catalog import Catalog
from pyiceberg.typedef import Identifier


class NamespaceTools:
    """Helper methods for working with Iceberg namespaces.

    Attributes:
        catalog: Catalog instance used to perform
            namespace operations.
    """

    catalog: Catalog

    def __init__(self, catalog: Catalog) -> None:
        """Initialize the NamespaceTools wrapper.

        Args:
            catalog: Catalog used for namespace
                operations.
        """
        self.catalog = catalog

    async def list_namespaces(
        self,
        namespace: Annotated[Union[str, Identifier], Field(description="Parent namespace identifier to search.")] = (),
    ) -> Annotated[List[Identifier], Field(description="List of namespace identifiers.")]:
        """List all namespaces under a parent namespace.

        Args:
            namespace: Parent namespace identifier to
                search. Defaults to the root namespace when omitted or empty.

        Returns:
            A list of namespace identifiers.
        """

        return self.catalog.list_namespaces(namespace)

    async def create_namespace(
        self,
        namespace: Annotated[Union[str, Identifier], Field(description="Namespace to create.")],
    ) -> Annotated[List[Identifier], Field(description="List of all namespaces under root namespace.")]:
        """Create a new namespace if it does not already exist.

        Args:
            namespace: Namespace to create.

        Returns:
            A list of all namespaces under the root namespace.
        """
        self.catalog.create_namespace_if_not_exists(namespace)

        return await self.list_namespaces()

    async def delete_namespace(
        self,
        namespace: Annotated[Union[str, Identifier], Field(description="Namespace to delete.")],
    ) -> Annotated[List[Identifier], Field(description="List of remaining namespaces under root namespace.")]:
        """Delete a namespace from the catalog.

        Args:
            namespace: Namespace to delete.

        Returns:
            List of remaining namespaces under root namespace.
        """
        self.catalog.drop_namespace(namespace)

        return await self.list_namespaces()



import tempfile
from pathlib import Path
from unittest import IsolatedAsyncioTestCase
from unittest.mock import Mock, patch

from duckdb import DuckDBPyConnection, DuckDBPyRelation
from polars import DataFrame
from pyiceberg.catalog import Catalog, CatalogType

from src.tools.query import QueryTools, load_duckdb


class TestQuerySQL(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.df = DataFrame({"a": [1, 2, 3], "b": [3, 4, 5]})
        self.mock_duckdb = Mock(spec=DuckDBPyConnection)
        self.tools = QueryTools(self.mock_duckdb)

        mock_result = Mock(spec=DuckDBPyRelation)
        mock_result.pl.return_value = self.df
        mock_result.arrow.return_value = self.df.to_arrow().to_reader()
        self.mock_duckdb.sql.return_value.execute.return_value = mock_result

    async def test_sql_query_with_csv_file(self) -> None:
        with tempfile.TemporaryDirectory() as parent_dir:
            file_path = Path(parent_dir) / "test.csv"

            await self.tools.sql_query("SELECT * FROM CATALOG", file_path)

            self.assertTrue(file_path.is_file())

    async def test_sql_query_with_unsupported_file_extension(self) -> None:
        with self.assertRaises(ValueError) as context:
            await self.tools.sql_query("SELECT * FROM CATALOG", Path("test.unsupported"))

        self.assertIn("Unsupported file extension", str(context.exception))

    async def test_sql_query_with_nonexistent_parent_directory(self) -> None:
        with self.assertRaises(FileNotFoundError) as context:
            await self.tools.sql_query("SELECT * FROM CATALOG", Path("/nonexistent/test.csv"))

        self.assertIn("Parent directory", str(context.exception))

    async def test_sql_query(self) -> None:
        result = await self.tools.sql_query("SELECT * FROM CATALOG")

        self.assertEqual(result, self.df.write_json())
        self.mock_duckdb.sql.assert_called_with("SELECT * FROM CATALOG")


class TestQueryLoadDuckDB(IsolatedAsyncioTestCase):
    @patch("src.tools.query.infer_catalog_type")
    @patch("src.tools.query.ddb_connect")
    def test_load_duckdb_with_rest_catalog_oauth2(
        self,
        mock_connect,
        mock_infer,
    ) -> None:
        mock_catalog = Mock(spec=Catalog)
        mock_catalog.name = "test_catalog"
        mock_catalog.properties = {
            "warehouse": "test_warehouse",
            "uri": "http://test-uri",
            "oauth2-server-uri": "http://oauth-uri",
            "client-id": "test-client-id",
            "client-secret": "test-client-secret",
        }

        mock_conn = Mock(spec=DuckDBPyConnection)
        mock_connect.return_value = mock_conn
        mock_infer.return_value = CatalogType.REST

        result = load_duckdb(mock_catalog)

        mock_conn.load_extension.assert_called_with("iceberg")
        self.assertEqual(result, mock_conn)

    @patch("src.tools.query.infer_catalog_type")
    @patch("src.tools.query.ddb_connect")
    def test_load_duckdb_with_rest_catalog_token(
        self,
        mock_connect,
        mock_infer,
    ) -> None:
        mock_catalog = Mock(spec=Catalog)
        mock_catalog.name = "test_catalog"
        mock_catalog.properties = {"warehouse": "test_warehouse", "uri": "http://test-uri", "token": "test-token"}

        mock_conn = Mock(spec=DuckDBPyConnection)
        mock_connect.return_value = mock_conn
        mock_infer.return_value = CatalogType.REST

        result = load_duckdb(mock_catalog)

        mock_conn.load_extension.assert_called_with("iceberg")
        self.assertEqual(result, mock_conn)

    @patch("src.tools.query.infer_catalog_type")
    @patch("src.tools.query.ddb_connect")
    def test_load_duckdb_with_unsupported_catalog_type(
        self,
        mock_connect,
        mock_infer,
    ) -> None:
        mock_catalog = Mock(spec=Catalog)
        mock_catalog.name = "test_catalog"
        mock_catalog.properties = {}

        mock_conn = Mock(spec=DuckDBPyConnection)
        mock_connect.return_value = mock_conn
        mock_infer.return_value = CatalogType.BIGQUERY  # Unsupported type

        result = load_duckdb(mock_catalog)

        self.assertIsNone(result)



import tempfile
from pathlib import Path
from unittest import IsolatedAsyncioTestCase
from unittest.mock import Mock

from polars import DataFrame
from pyiceberg.catalog import Catalog
from pyiceberg.table import Table
from pyiceberg.table.metadata import TableMetadata
from pyiceberg.table.snapshots import Snapshot

from src.tools.table import TableTools


class TestTable(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.mock_catalog = Mock(spec=Catalog)
        self.tools = TableTools(self.mock_catalog)
        self.mock_table = Mock(spec=Table)
        self.mock_catalog.load_table.return_value = self.mock_table

        self.mock_metadata = Mock(spec=TableMetadata)
        self.mock_table.metadata = self.mock_metadata

        self.mock_snapshot1 = Mock(spec=Snapshot)
        self.mock_snapshot1.snapshot_id = 1

        self.mock_snapshot2 = Mock(spec=Snapshot)
        self.mock_snapshot2.snapshot_id = 2
        self.mock_snapshot2.parent_snapshot_id = 1

        self.mock_table.current_snapshot.return_value = self.mock_snapshot2
        self.mock_table.inspect.snapshots.return_value = [self.mock_snapshot1, self.mock_snapshot2]

        self.df = DataFrame({"a": [1, 2, 3], "b": [3, 4, 5]})
        self.mock_snapshot2.summary = {"total-records": len(self.df)}
        self.mock_table.to_polars.side_effect = OSError
        self.mock_table.scan.return_value.to_polars.return_value = self.df

    async def test_list_tables_returns_table_list(self) -> None:
        self.mock_catalog.list_tables.return_value = [("table1",), ("table2",)]

        result = await self.tools.list_tables("namespace")

        self.assertEqual(result, [("table1",), ("table2",)])
        self.assertEqual(self.mock_catalog.list_tables.call_args, (("namespace",),))
        self.assertEqual(self.mock_catalog.list_tables.call_count, 1)

    async def test_read_table_metadata_returns_table_metadata(self) -> None:
        result = await self.tools.read_table_metadata("test_table")

        self.assertEqual(result, self.mock_metadata)

        self.assertEqual(self.mock_catalog.load_table.call_args, (("test_table",),))
        self.assertEqual(self.mock_catalog.load_table.call_count, 1)

    async def test_read_table_contents_returns_full_table(self) -> None:
        result = await self.tools.read_table_contents("test_table")

        self.assertEqual(result, self.df.write_json())

    async def test_read_table_contents_with_pagination_returns_page(self) -> None:
        result = await self.tools.read_table_contents("test_table", end=2)

        self.assertEqual(result, DataFrame({"a": [1, 2], "b": [3, 4]}).write_json())

    async def test_create_table_returns_tail(self) -> None:
        self.mock_catalog.create_table.return_value = self.mock_table

        await self.tools.create_table("test_table", self.df.to_dict(as_series=False))

        self.mock_catalog.create_table.assert_called_with("test_table", self.df.to_arrow().schema)
        self.mock_table.overwrite.assert_called_with(self.df.to_arrow())

    async def test_overwrite_table_returns_tail(self) -> None:
        self.mock_catalog.create_table.return_value = self.mock_table

        await self.tools.write_table("test_table", "overwrite", self.df.to_dict(as_series=False))

        self.mock_table.overwrite.assert_called_with(self.df.to_arrow())

    async def test_append_table_returns_tail(self) -> None:
        self.mock_catalog.create_table.return_value = self.mock_table

        await self.tools.write_table("test_table", "append", self.df.to_dict(as_series=False))

        self.mock_table.append.assert_called_with(self.df.to_arrow())

    async def test_read_table_snapshots_returns_all_snapshots(self) -> None:
        result = await self.tools.read_table_snapshots("test_table")

        self.assertEqual(len(result), 2)

        self.assertEqual(result[0].snapshot_id, 1)
        self.assertEqual(result[1].snapshot_id, 2)

    async def test_read_table_snapshots_filters_by_snapshot_id(self) -> None:
        result = await self.tools.read_table_snapshots("test_table", snapshot_id=1)

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].snapshot_id, 1)

    async def test_read_table_snapshots_respects_limit(self) -> None:
        result = await self.tools.read_table_snapshots("test_table", limit=1)

        self.assertEqual(len(result), 1)
        self.assertEqual(result[0].snapshot_id, 1)

    async def test_delete_table_calls_drop_and_returns_remaining_tables(self) -> None:
        self.mock_catalog.drop_table.return_value = None
        self.mock_catalog.list_tables.return_value = [("remaining_table",)]

        result = await self.tools.delete_table("test_table")

        self.assertEqual(self.mock_catalog.drop_table.call_args, (("test_table",),))
        self.assertEqual(self.mock_catalog.drop_table.call_count, 1)
        self.assertEqual(result, [("remaining_table",)])

    async def test_download_table_contents_with_csv_file(self) -> None:
        with tempfile.TemporaryDirectory() as parent_dir:
            file_path = Path(parent_dir) / "test.csv"

            self.mock_table.scan.return_value.to_arrow_batch_reader.return_value = self.df.to_arrow().to_reader()

            await self.tools.download_table_contents("test_table", file_path)

            self.assertTrue(file_path.is_file())

    async def test_download_table_contents_with_unsupported_file_extension(self) -> None:
        with self.assertRaises(ValueError):
            await self.tools.download_table_contents("test_table", Path("test.unsupported"))

    async def test_download_table_contents_with_nonexistent_parent_directory(self) -> None:
        with self.assertRaises(FileNotFoundError):
            await self.tools.download_table_contents("test_table", Path("/nonexistent/test.csv"))

    async def test_create_table_with_contents(self) -> None:
        self.mock_catalog.create_table.return_value = self.mock_table

        result = await self.tools.create_table("test_table", self.df.to_dict(as_series=False))

        self.mock_catalog.create_table.assert_called_once()
        self.mock_table.overwrite.assert_called_once()
        self.assertIsInstance(result, str)

    async def test_create_table_with_file(self) -> None:
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = Path(temp_dir) / "test.csv"
            self.df.write_csv(file_path)

            self.mock_catalog.create_table.return_value = self.mock_table

            await self.tools.create_table("test_table", file=file_path)

            self.mock_catalog.create_table.assert_called_once()
            self.mock_table.overwrite.assert_called_once()

    async def test_create_table_with_both_contents_and_file(self) -> None:
        with self.assertRaises(ValueError):
            await self.tools.create_table("test_table", self.df.to_dict(as_series=False), file=Path("test.csv"))

    async def test_create_table_with_neither_contents_nor_file(self) -> None:
        with self.assertRaises(ValueError):
            await self.tools.create_table("test_table")

    async def test_write_table_append_mode(self) -> None:
        await self.tools.write_table("test_table", "append", self.df.to_dict(as_series=False))

        self.mock_table.append.assert_called_once()

    async def test_write_table_overwrite_mode(self) -> None:
        await self.tools.write_table("test_table", "overwrite", self.df.to_dict(as_series=False))

        self.mock_table.overwrite.assert_called_once()

    async def test_write_table_with_file(self) -> None:
        with tempfile.TemporaryDirectory() as temp_dir:
            file_path = Path(temp_dir) / "test.csv"
            self.df.write_csv(file_path)

            await self.tools.write_table("test_table", "append", file=file_path)

            self.mock_table.append.assert_called_once()

    async def test_write_table_with_both_contents_and_file(self) -> None:
        with self.assertRaises(ValueError) as context:
            await self.tools.write_table(
                "test_table", "append", self.df.to_dict(as_series=False), file=Path("test.csv")
            )

        self.assertIn("Only one of contents or file can be provided", str(context.exception))

    async def test_write_table_with_neither_contents_nor_file(self) -> None:
        with self.assertRaises(ValueError) as context:
            await self.tools.write_table("test_table", "append")

        self.assertIn("One of contents or file must be provided", str(context.exception))



from os import getenv

from fastmcp import FastMCP
from mcp.types import ToolAnnotations
from pyiceberg.catalog import load_catalog
from sentry_sdk import init as sentry_init
from sentry_sdk.integrations.mcp import MCPIntegration

from src.tools.namespace import NamespaceTools
from src.tools.query import QueryTools, load_duckdb
from src.tools.table import TableTools

catalog = load_catalog(getenv("ICEBERG_CATALOG"))
duckdb = load_duckdb(catalog)
mcp = FastMCP(
    name="Iceberg MCP Server",
)

namespace = NamespaceTools(catalog)
mcp.tool(namespace.list_namespaces, annotations=ToolAnnotations(readOnlyHint=True))
mcp.tool(namespace.create_namespace)
mcp.tool(namespace.delete_namespace, annotations=ToolAnnotations(destructiveHint=True))

table = TableTools(catalog)
mcp.tool(table.list_tables, annotations=ToolAnnotations(readOnlyHint=True))
mcp.tool(table.read_table_metadata, annotations=ToolAnnotations(readOnlyHint=True))
mcp.tool(table.read_table_contents, annotations=ToolAnnotations(readOnlyHint=True))
mcp.tool(table.download_table_contents, task=True)
mcp.tool(table.read_table_snapshots, annotations=ToolAnnotations(readOnlyHint=True))
mcp.tool(table.create_table)
mcp.tool(table.write_table)
mcp.tool(table.delete_table, annotations=ToolAnnotations(destructiveHint=True))

if duckdb is not None:
    query = QueryTools(duckdb)
    mcp.tool(query.sql_query, annotations=ToolAnnotations(destructiveHint=True))

if getenv("SENTRY_DSN") is not None:
    sentry_init(
        dsn=getenv("SENTRY_DSN"),
        enable_tracing=True,
        traces_sample_rate=1.0,
        profiles_sample_rate=1.0,
        enable_logs=True,
        send_default_pii=True,
        integrations=[MCPIntegration()],
    )


if __name__ == "__main__":
    mcp.run()
